ALL = "All tested clients"

OUTLOOK = "Outlook 2007/10/13"
OUTLOOK_EXPRESS = "Outlook 03/Express/Mail"
OUTLOOK_COM = "Outlook.com"

APPLE_MAIL = "Apple Mail 6.5"
IPHONE = "Iphone iOS 7/iPad"

YAHOO = "Yahoo! Mail"

GMAIL = "Google Gmail"

ANDROID = "Android 4 (Gmail)"



from clients import OUTLOOK, OUTLOOK_EXPRESS, IPHONE, OUTLOOK_COM, APPLE_MAIL, YAHOO, GMAIL, ANDROID

# https://www.campaignmonitor.com/css/
INVALID_PROPERTIES = {
    # text & font
    "text-overflow": [OUTLOOK],
    "text-shadow": [OUTLOOK, OUTLOOK_EXPRESS, GMAIL],
    "word-wrap": [OUTLOOK, GMAIL, ANDROID],
    "text-fill-color": [OUTLOOK, OUTLOOK_EXPRESS, OUTLOOK_COM, YAHOO, GMAIL],
    "text-fill-stroke": [OUTLOOK, OUTLOOK_EXPRESS, OUTLOOK_COM, YAHOO, GMAIL],
    # color & background
    "background": [OUTLOOK, OUTLOOK_EXPRESS, OUTLOOK_COM, YAHOO, GMAIL],
    "background-image": [OUTLOOK, OUTLOOK_COM],
    "background-position": [OUTLOOK, OUTLOOK_COM, GMAIL],
    "background-repeat": [OUTLOOK, OUTLOOK_COM],
    "background-size": [OUTLOOK, OUTLOOK_EXPRESS, OUTLOOK_COM, YAHOO, GMAIL, ANDROID],
    # box model
    "border-color": [OUTLOOK, OUTLOOK_EXPRESS, ANDROID],
    "border-image": [OUTLOOK, OUTLOOK_EXPRESS, OUTLOOK_COM, YAHOO, GMAIL, ANDROID],
    "border-radius": [OUTLOOK, OUTLOOK_EXPRESS, YAHOO, ANDROID],
    "box-shadow": [OUTLOOK, OUTLOOK_EXPRESS, YAHOO, GMAIL, ANDROID],
    "margin": [OUTLOOK_COM],
    "max-width": [OUTLOOK],
    "min-width": [OUTLOOK],
    # positioning & display
    "bottom": [OUTLOOK, YAHOO, GMAIL],
    "clear": [OUTLOOK],
    "clip": [OUTLOOK, OUTLOOK_COM, YAHOO, GMAIL],
    "cursor": [OUTLOOK, GMAIL, ANDROID],
    "display": [OUTLOOK, GMAIL],
    "float": [OUTLOOK, OUTLOOK_COM],
    "left": [OUTLOOK, OUTLOOK_COM, YAHOO, GMAIL],
    "opacity": [OUTLOOK, OUTLOOK_EXPRESS, YAHOO, GMAIL],
    "outline": [OUTLOOK, OUTLOOK_EXPRESS],
    "overflow": [OUTLOOK],
    "position": [OUTLOOK, YAHOO, GMAIL],
    "resize": [OUTLOOK, GMAIL, ANDROID],
    "right": [OUTLOOK, YAHOO, GMAIL],
    "top": [OUTLOOK, YAHOO, GMAIL],
    "visibility": [OUTLOOK, GMAIL],
    "z-index": [GMAIL],
    # lists
    "list-style-image": [OUTLOOK, OUTLOOK_COM, GMAIL],
    "list-style-position": [OUTLOOK, OUTLOOK_COM],
    # tables
    "border-spacing": [OUTLOOK, OUTLOOK_EXPRESS],
    "caption-side": [OUTLOOK, OUTLOOK_EXPRESS],
    "empty-cells": [OUTLOOK, OUTLOOK_EXPRESS],
}


import tinycss

from lxml import etree
from lxml.cssselect import CSSSelector
from lxml.html import soupparser

from linter import get_clients_without_support


# specificity is represented as a single int at the moment
# so to account for !important we need to make a big int
IMPORTANT_MULTIPLIER = 9000

def inline_css(html, css, strip_unsupported_css=True, remove_classes=False, pretty_print=True):
    """Produces html output with css styles inlined.

    :param html: the original html
    :param css: the css rules to apply to the html
    :param strip_unsupported_css: a flag for stripping invalid email css
    :param remove_classes: a flag for stripping class attributes from the output
    :param pretty_print: toggle pretty printing of resulting html
    """
    node_declarations = {}
    try:
        dom = etree.fromstring(html)
    except etree.XMLSyntaxError:
        # fall back on the soup parser if we don't have valid XML
        dom = soupparser.fromstring(html)

    css_rules = tinycss.make_parser().parse_stylesheet(css).rules
    for rule in css_rules:
        css_selector = rule.selector.as_css()
        for node in CSSSelector(css_selector)(dom):
            for declaration in rule.declarations:

                if strip_unsupported_css and get_clients_without_support(declaration):
                    # skip invalid rules
                    continue

                declaration.specificity = _selector_specificity(css_selector, declaration.priority)
                node_declarations.setdefault(node, {})
                active_declaration = node_declarations[node].get(declaration.name)
                if active_declaration and active_declaration.specificity > declaration.specificity:
                    # skip rules with lower specificity
                    continue

                node_declarations[node][declaration.name] = declaration

    for node, declarations in node_declarations.iteritems():
        inline_styles = node.get('style', '')
        inline_css = _get_node_style(declarations, inline_styles)
        node.set('style', inline_css)
        if remove_classes and "class" in node.attrib:
            node.attrib.pop('class')

    return etree.tostring(dom, pretty_print=pretty_print)


def _get_node_style(declarations, inline_styles):
    """Given css and inline styles determine the value of the style attribute."""
    inline_styles = _parse_style_attribute(inline_styles)
    stringify_value = lambda value: " ".join([v.as_css() for v in value])
    style = " ".join(["{}: {};".format(declaration.name, stringify_value(declaration.value)) \
                     for declaration in declarations.values() \
                     if declaration.name not in inline_styles])
    if inline_styles:
        if style:
            style = style + " "
        style = style + " ".join(["{}: {};".format(name, value) \
                                        for (name, value) in inline_styles.iteritems()])
    return style

def _parse_style_attribute(inline_styles):
    """Parse the style attribute value into a dict of css declarations."""
    inline_css = {}
    inline_styles = inline_styles.split(";")
    for rule in inline_styles:
        rule = rule.split(":")
        if len(rule) != 2:
            continue
        name, value = rule
        name = name.strip()
        value = value.strip()
        if name and value:
            inline_css[name] = value
    return inline_css

def _selector_specificity(selector, priority):
    """Calculate an integer representing selector specificity.

    The higher the number the more specific. This is a simplification of the
    spec, but works pretty well in practice.

    :param selector: the css selector
    :param priority: boolean that is set if the rule has the `!important` flag.
    """
    class_weight = selector.count(".")
    id_weight = selector.count("#")
    element_weight = len([a for a in selector.split(" ") if not (a.startswith(".") or a.startswith("#"))])
    weight = element_weight + class_weight * 10 + element_weight * id_weight * 100
    if priority:
        weight *= IMPORTANT_MULTIPLIER
    return weight



import unittest

from emailipy.inliner import inline_css

class TestInliner(unittest.TestCase):
    def test_noop(self):
        """Verify html remains unchanged with no css passed"""
        html = '<div class="pink">test</div>'
        css = ''
        self.assertEqual(html, inline_css(html, css, pretty_print=False))

    def test_class_rule(self):
        """Verify classes are applied"""
        html = '<div class="pink">test</div>'
        css = '.pink { font-size: 1em; }'
        expected = '<div class="pink" style="font-size: 1em;">test</div>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_class_no_class(self):
        """Verify classes are applied only where intended"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = '.pink { font-size: 1em; }'
        expected = '<span><div class="pink" style="font-size: 1em;">test</div><div>t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_classes_stripped(self):
        """Verify classes are stripped w/ remove_classes flag."""
        html = '<div class="pink">test</div>'
        css = '.pink { font-size: 1em; }'
        expected = '<div style="font-size: 1em;">test</div>'
        result = inline_css(html, css, remove_classes=True, pretty_print=False)
        self.assertEqual(expected, result)

    def test_element_rule(self):
        """Verify element rules are applied"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = 'div { font-size: 1em; }'
        expected = '<span><div class="pink" style="font-size: 1em;">test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_css3_nth_child(self):
        """Check css3 nth-child support"""
        html = '<span><div>test</div><div>t2</div></span>'
        css = 'div:nth-child(2) { font-size: 1em; }'
        expected = '<span><div>test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_class_element_specificity(self):
        """Verify class rules override element rules"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = 'div { font-size: 1em; } .pink { font-size: 2em; }'
        expected = '<span><div class="pink" style="font-size: 2em;">test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_important(self):
        """Verify `!important` is respected"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = 'div { font-size: 1em !important; } .pink { font-size: 2em; }'
        expected = '<span><div class="pink" style="font-size: 1em;">test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_inline_styles(self):
        """Check if inline styles win the specificity war"""
        html = '<span><div class="pink" style="font-size: 1em;">test</div><div>t2</div></span>'
        css = 'div { font-size: 3em; } .pink { font-size: 2em; }'
        expected = '<span><div class="pink" style="font-size: 1em;">test</div><div style="font-size: 3em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_invalid_rule(self):
        """Verify email styles are stripped"""
        html = '<div class="pink">test</div>'
        css = '.pink { opacity: 0.8; }'
        expected = '<div class="pink">test</div>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_invalid_rule_not_stripped(self):
        """Make sure we can turn off unsupported css stripping"""
        html = '<div class="pink">test</div>'
        css = '.pink { opacity: 0.8; }'
        expected = '<div class="pink" style="opacity: 0.8;">test</div>'
        result = inline_css(html, css, strip_unsupported_css=False, pretty_print=False)
        self.assertEqual(expected, result)

    def test_id_rule(self):
        """Verify classes are applied"""
        html = '<div id="pink">test</div>'
        css = '#pink { font-size: 1em; }'
        expected = '<div id="pink" style="font-size: 1em;">test</div>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_wildcard_rule(self):
        """Verify element rules are applied"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = 'span * { font-size: 1em; }'
        expected = '<span><div class="pink" style="font-size: 1em;">test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_child_rule(self):
        """Verify element rules are applied"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = 'span > div { font-size: 1em; }'
        expected = '<span><div class="pink" style="font-size: 1em;">test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)

    def test_sibling_rule(self):
        """Verify element rules are applied"""
        html = '<span><div class="pink">test</div><div>t2</div></span>'
        css = 'span > div.pink ~ * { font-size: 1em; }'
        expected = '<span><div class="pink">test</div><div style="font-size: 1em;">t2</div></span>'
        result = inline_css(html, css, pretty_print=False)
        self.assertEqual(expected, result)



import click
import os
import sys

from linter import lint_css
from inliner import inline_css


@click.command()
@click.argument('css_file', required=True, type=click.File('r'))
def lint(css_file):
    """Lints email css and prints issues per client."""
    css = css_file.read()

    issues = lint_css(css) or []
    for issue in issues:
        click.echo(issue)


@click.command()
@click.option('--allow_invalid_css', '-i', is_flag=True, help="Allows css that doesn't lint.")
@click.option('--remove_classes', '-c', is_flag=True, help="Strip all class attributes after inlining")
@click.argument('html_file', required=True, type=click.File('r'))
@click.argument('css_file', required=True, type=click.File('r'))
def inline(html_file, css_file, allow_invalid_css, remove_classes):
    """Inlines css into html to make it safe for email."""
    files = {}
    for extension, f in [("html", html_file), ("css", css_file)]:
        files[extension] = f.read()

    html = inline_css(*files.values(),
                      strip_unsupported_css=(not allow_invalid_css),
                      remove_classes=remove_classes)
    click.echo(html)



import tinycss
import rules


def lint_css(css):
    violations = []
    css_rules = tinycss.make_parser().parse_stylesheet(css).rules
    for rule in css_rules:
        for declaration in rule.declarations:
            css_selector = rule.selector.as_css()
            prop = declaration.name
            value = " ".join([v.as_css() for v in declaration.value])
            clients_without_support = get_clients_without_support(declaration)
            if clients_without_support:
                clients = " | ".join(clients_without_support)
                warning = "Invalid Rule: {} {{ {}: {}; }} -- {}".format(css_selector, prop, value, clients)

                violations.append(warning)

    return violations

# uses https://www.campaignmonitor.com/css/
def get_clients_without_support(declaration):
    return rules.INVALID_PROPERTIES.get(declaration.name)



import sys

from setuptools import setup, find_packages


install_requires = [l.split('#')[0].strip()
                    for l in open('requirements.txt').readlines()
                    if not l.startswith('#') and not l.startswith('-e')]

setup(
    name='emailipy',
    packages = find_packages(),
    version='1.0.0',
    url='https://github.com/Parsely/emailipy',
    description='Inlines css into html to make it safe for email.',
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'emailipy-lint = emailipy.cli:lint',
            'emailipy-inline = emailipy.cli:inline',
        ],
    },
    )
